@@ -12,6 +12,8 @@
  * Department of Chemical Engineering
  * The University of Texas at Austin
  * Austin, Texas  78712
+ *
+ * Some manpath, compression and locale related changes - aeb - 940320
  */
 
 #define MAN_MAIN
@@ -21,11 +23,12 @@
 #include <string.h>
 #include <sys/file.h>
 #include <signal.h>
-#include "config.h"
 #include "gripes.h"
 #include "version.h"
+#include "config.h"
+#include "man.h"
 
-#ifndef POSIX
+#ifdef POSIX
 #include <unistd.h>
 #else
 #ifndef R_OK
@@ -56,8 +59,8 @@ extern int fclose ();
 extern char *sprintf ();
 #endif
 
-extern char *strdup ();
-
+extern char *get_expander ();
+extern char *getval ();
 extern char **glob_vector ();
 extern char **glob_filename ();
 extern int access ();
@@ -69,40 +72,21 @@ extern int is_directory ();
 extern int do_system_command ();
 
 char *prognam;
-static char *pager;
-static char *manp;
-static char *manpathlist[MAXDIRS];
-static char *section;
-static char *colon_sep_section_list;
-static char **section_list;
-static char *roff_directive;
-static int apropos;
-static int whatis;
-static int findall;
-static int print_where;
-
-#ifdef ALT_SYSTEMS
-static int alt_system;
-static char *alt_system_name;
-#endif
-
-static int troff = 0;
-
+char *pager;
+char *colon_sep_section_list;
+char *roff_directive;
+int apropos;
+int whatis;
+int findall;
+int print_where;
+int do_troff;
 int debug;
+struct dir srchlist;
 
-#ifdef HAS_TROFF
-#ifdef ALT_SYSTEMS
-static char args[] = "M:P:S:adfhkm:p:tw?";
-#else
-static char args[] = "M:P:S:adfhkp:tw?";
-#endif
-#else
-#ifdef ALT_SYSTEMS
-static char args[] = "M:P:S:adfhkm:p:w?";
-#else
-static char args[] = "M:P:S:adfhkp:w?";
-#endif
-#endif
+static char *section;
+static char **section_list;
+
+static int have_roff = 1;
 
 int
 main (argc, argv)
@@ -114,6 +98,7 @@ main (argc, argv)
   char *tmp;
   extern int optind;
   extern char *mkprogname ();
+  extern void prmanpath();
   char *is_section ();
   char **get_section_list ();
   void man_getopt ();
@@ -125,6 +110,11 @@ main (argc, argv)
 
   man_getopt (argc, argv);
 
+  if (!strcmp (prognam, "manpath") || (optind == argc && print_where)) {
+      prmanpath();
+      exit(0);
+  }
+
   if (optind == argc)
     gripe_no_name ((char *)NULL);
 
@@ -178,7 +168,6 @@ usage ()
 {
   static char usage_string[1024] = "%s, version %s\n\n";
 
-#ifdef HAS_TROFF
 #ifdef ALT_SYSTEMS
   static char s1[] =
     "usage: %s [-adfhktw] [section] [-M path] [-P pager] [-S list]\n\
@@ -188,17 +177,6 @@ usage ()
     "usage: %s [-adfhktw] [section] [-M path] [-P pager] [-S list]\n\
            [-p string] name ...\n\n";
 #endif
-#else
-#ifdef ALT_SYSTEMS
-  static char s1[] =
-    "usage: %s [-adfhkw] [section] [-M path] [-P pager] [-S list]\n\
-           [-m system] [-p string] name ...\n\n";
-#else
-  static char s1[] =
-    "usage: %s [-adfhkw] [section] [-M path] [-P pager] [-S list]\n\
-           [-p string] name ...\n\n";
-#endif
-#endif
 
 static char s2[] = "  a : find all matching entries\n\
   d : print gobs of debugging information\n\
@@ -206,11 +184,11 @@ static char s2[] = "  a : find all matching entries\n\
   h : print this help message\n\
   k : same as apropos(1)\n";
 
-#ifdef HAS_TROFF
   static char s3[] = "  t : use troff to format pages for printing\n";
-#endif
 
-  static char s4[] = "  w : print location of man page(s) that would be displayed\n\n\
+  static char s4[] = "\
+  w : print location of man page(s) that would be displayed\n\
+      (if no name given: print directories that would be searched)\n\n\
   M path   : set search path for manual pages to `path'\n\
   P pager  : use program `pager' to display pages\n\
   S list   : colon separated section list\n";
@@ -225,11 +203,7 @@ static char s2[] = "  a : find all matching entries\n\
 
   strcat (usage_string, s1);
   strcat (usage_string, s2);
-
-#ifdef HAS_TROFF
-  strcat (usage_string, s3);
-#endif
-
+  strcat (usage_string, s3);    /* maybe only if troff found? */
   strcat (usage_string, s4);
 
 #ifdef ALT_SYSTEMS
@@ -242,190 +216,6 @@ static char s2[] = "  a : find all matching entries\n\
   exit(1);
 }
 
-char **
-add_dir_to_mpath_list (mp, p)
-     char **mp;
-     char *p;
-{
-  int status;
-
-  status = is_directory (p);
-
-  if (status < 0)
-    {
-      fprintf (stderr, "Warning: couldn't stat file %s!\n", p);
-    }
-  else if (status == 0)
-    {
-      fprintf (stderr, "Warning: %s isn't a directory!\n", p);
-    }
-  else if (status == 1)
-    {
-      if (debug)
-	fprintf (stderr, "adding %s to manpathlist\n", p);
-
-      *mp++ = strdup (p);
-    }
-  return mp;
-}
-
-/*
- * Get options from the command line and user environment.
- */
-void
-man_getopt (argc, argv)
-     register int argc;
-     register char **argv;
-{
-  register int c;
-  register char *p;
-  register char *end;
-  register char **mp;
-  extern char *optarg;
-  extern int getopt ();
-  extern void downcase ();
-  extern char *manpath ();
-
-  while ((c = getopt (argc, argv, args)) != EOF)
-    {
-      switch (c)
-	{
-	case 'M':
-	  manp = strdup (optarg);
-	  break;
-	case 'P':
-	  pager = strdup (optarg);
-	  break;
-	case 'S':
-	  colon_sep_section_list = strdup (optarg); 
-	  break;
-	case 'a':
-	  findall++;
-	  break;
-	case 'd':
-	  debug++;
-	  break;
-	case 'f':
-	  if (troff)
-	    gripe_incompatible ("-f and -t");
-	  if (apropos)
-	    gripe_incompatible ("-f and -k");
-	  if (print_where)
-	    gripe_incompatible ("-f and -w");
-	  whatis++;
-	  break;
-	case 'k':
-	  if (troff)
-	    gripe_incompatible ("-k and -t");
-	  if (whatis)
-	    gripe_incompatible ("-k and -f");
-	  if (print_where)
-	    gripe_incompatible ("-k and -w");
-	  apropos++;
-	  break;
-#ifdef ALT_SYSTEMS
-	case 'm':
-	  alt_system++;
-	  alt_system_name = strdup (optarg);
-	  break;
-#endif
-	case 'p':
-	  roff_directive = strdup (optarg);
-	  break;
-#ifdef HAS_TROFF
-	case 't':
-	  if (apropos)
-	    gripe_incompatible ("-t and -k");
-	  if (whatis)
-	    gripe_incompatible ("-t and -f");
-	  if (print_where)
-	    gripe_incompatible ("-t and -w");
-	  troff++;
-	  break;
-#endif
-	case 'w':
-	  if (apropos)
-	    gripe_incompatible ("-w and -k");
-	  if (whatis)
-	    gripe_incompatible ("-w and -f");
-	  if (troff)
-	    gripe_incompatible ("-w and -t");
-	  print_where++;
-	  break;
-	case 'h':
-	case '?':
-	default:
-	  usage();
-	  break;
-	}
-    }
-
-  if (pager == NULL || *pager == '\0')
-    if ((pager = getenv ("PAGER")) == NULL)
-      pager = strdup (PAGER);
-
-  if (debug)
-    fprintf (stderr, "\nusing %s as pager\n", pager);
-
-  if (manp == NULL)
-    {
-      if ((manp = manpath (0)) == NULL)
-	gripe_manpath ();
-
-      if (debug)
-	fprintf (stderr,
-		 "\nsearch path for pages determined by manpath is\n%s\n\n",
-		 manp);
-    }
-
-#ifdef ALT_SYSTEMS
-  if (alt_system_name == NULL || *alt_system_name == '\0')
-    if ((alt_system_name = getenv ("SYSTEM")) != NULL)
-      alt_system_name = strdup (alt_system_name);
-
-  if (alt_system_name != NULL && *alt_system_name != '\0')
-    downcase (alt_system_name);
-#endif
-
-  /*
-   * Expand the manpath into a list for easier handling.
-   */
-  mp = manpathlist;
-  for (p = manp; ; p = end+1)
-    {
-      if ((end = strchr (p, ':')) != NULL)
-	*end = '\0';
-
-#ifdef ALT_SYSTEMS
-      if (alt_system)
-	{
-	  char buf[BUFSIZ];
-
-	  if (debug)
-	    fprintf (stderr, "Alternate system `%s' specified\n",
-		     alt_system_name);
-
-	  strcpy (buf, p);
-	  strcat (buf, "/");
-	  strcat (buf, alt_system_name);
-
-	  mp = add_dir_to_mpath_list (mp, buf);
-	}
-      else
-	{
-	  mp = add_dir_to_mpath_list (mp, p);
-	}
-#else
-      mp = add_dir_to_mpath_list (mp, p);
-#endif
-      if (end == NULL)
-	break;
-
-      *end = ':';
-    }
-  *mp = NULL;
-}
-
 /*
  * Check to see if the argument is a valid section number.  If the
  * first character of name is a numeral, or the name matches one of
@@ -440,7 +230,8 @@ is_section (name)
   register char **vs;
 
   for (vs = section_list; *vs != NULL; vs++)
-    if ((strcmp (*vs, name) == NULL) || (isdigit (name[0])))
+    if ((strcmp (*vs, name) == 0)
+	|| (isdigit (name[0]) && !isdigit (name[1])))
       return strdup (name);
 
   return NULL;
@@ -456,12 +247,12 @@ do_apropos (name)
   register int len;
   register char *command;
 
-  len = strlen (APROPOS) + strlen (name) + 2;
+  len = strlen (getval("APROPOS")) + strlen (name) + 2;
 
   if ((command = (char *) malloc(len)) == NULL)
     gripe_alloc (len, "command");
 
-  sprintf (command, "%s %s", APROPOS, name);
+  sprintf (command, "%s %s", getval("APROPOS"), name);
 
   (void) do_system_command (command);
 
@@ -478,12 +269,12 @@ do_whatis (name)
   register int len;
   register char *command;
 
-  len = strlen (WHATIS) + strlen (name) + 2;
+  len = strlen (getval("WHATIS")) + strlen (name) + 2;
 
   if ((command = (char *) malloc(len)) == NULL)
     gripe_alloc (len, "command");
 
-  sprintf (command, "%s %s", WHATIS, name);
+  sprintf (command, "%s %s", getval("WHATIS"), name);
 
   (void) do_system_command (command);
 
@@ -491,30 +282,23 @@ do_whatis (name)
 }
 
 /*
- * Change a name of the form ...man/man1/name.1 to ...man/cat1/name.1
- * or a name of the form ...man/cat1/name.1 to ...man/man1/name.1
+ * Change a name of the form ...man/man1/name.1 to ...man/cat1/name.1.gz
  */
 char *
-convert_name (name, to_cat)
+convert_name (name)
      register char *name;
-     register int to_cat;
 {
   register char *to_name;
   register char *t1;
   register char *t2 = NULL;
 
 #ifdef DO_COMPRESS
-  if (to_cat)
-    {
-      int len = strlen (name) + 3;
-      to_name = (char *) malloc (len);
-      if (to_name == NULL)
-	gripe_alloc (len, "to_name");
-      strcpy (to_name, name);
-      strcat (to_name, ".Z");
-    }
-  else
-    to_name = strdup (name);
+  int len = strlen (name) + strlen (getval("COMPRESS_EXT")) + 1;
+  to_name = (char *) malloc (len);
+  if (to_name == NULL)
+    gripe_alloc (len, "to_name");
+  strcpy (to_name, name);
+  strcat (to_name, getval("COMPRESS_EXT"));
 #else
   to_name = strdup (name);
 #endif
@@ -522,24 +306,16 @@ convert_name (name, to_cat)
   t1 = strrchr (to_name, '/');
   if (t1 != NULL)
     {
-      *t1 = NULL;
+      *t1 = 0;
       t2 = strrchr (to_name, '/');
       *t1 = '/';
     }
 
   if (t2 == NULL)
-    gripe_converting_name (name, to_cat);
+    gripe_converting_name (name);
 
-  if (to_cat)
-    {
-      *(++t2) = 'c';
-      *(t2+2) = 't';
-    }
-  else
-    {
-      *(++t2) = 'm';
-      *(t2+2) = 'n';
-    }
+  t2[1] = 'c';
+  t2[3] = 't';
 
   if (debug)
     fprintf (stderr, "to_name in convert_name () is: %s\n", to_name);
@@ -547,6 +323,61 @@ convert_name (name, to_cat)
   return to_name;
 }
 
+remove_file (file)
+     char *file;
+{
+  int i;
+
+  i = unlink (file);
+
+  if (debug) {
+      if (i)
+	perror(file);
+      else
+	fprintf (stderr, "unlinked %s\n", file);
+  }
+}
+
+remove_other_catfiles (catfile)
+     char *catfile;
+{
+#ifdef DO_COMPRESS
+  char pathname[BUFSIZ];
+  char *t;
+  char **gf;
+  int i, offset;
+
+  strcpy (pathname, catfile);
+  t = rindex(pathname, '.');
+  if (t == NULL || strcmp(t, getval("COMPRESS_EXT")))
+    return;
+  offset = t - pathname;
+  strcpy(t, "*");
+  gf = glob_filename (pathname);
+
+  if (gf != (char **) -1 && gf != NULL) {
+      for ( ; *gf; gf++) {
+	  /*
+	   * Only remove files with a known extension, like .Z
+	   * (otherwise we might kill a lot when called with
+	   * catfile = ".gz" ...)
+	   */
+	  if (strlen (*gf) <= offset) {
+	      if (strlen (*gf) == offset)  /* uncompressed version */
+		  remove_file (*gf);
+	      continue;
+	  }
+
+	  if (!strcmp (*gf + offset, getval("COMPRESS_EXT")))
+	    continue;
+
+	  if (get_expander (*gf) != NULL)
+	    remove_file (*gf);
+      }
+  }
+#endif
+}
+
 /*
  * Try to find the man page corresponding to the given name.  The
  * reason we do this with globbing is because some systems have man
@@ -554,8 +385,6 @@ convert_name (name, to_cat)
  * XtPopup.3Xt.  Rather than requiring that this program know about
  * all those possible names, we simply try to match things like
  * .../man[sect]/name[sect]*.  This is *much* easier.
- *
- * Note that globbing is only done when the section is unspecified.
  */
 char **
 glob_for_file (path, section, name, cat)
@@ -577,7 +406,7 @@ glob_for_file (path, section, name, cat)
 
   gf = glob_filename (pathname);
 
-  if ((gf == (char **) -1 || *gf == NULL) && isdigit (*section))
+  if ((gf == (char **) -1 || gf == NULL || *gf == NULL) && isdigit (*section))
     {
       if (cat)
 	sprintf (pathname, "%s/cat%s/%s.%c*", path, section, name, *section);
@@ -586,91 +415,11 @@ glob_for_file (path, section, name, cat)
 
       gf = glob_filename (pathname);
     }
+  if (gf == (char **) -1)
+    gf = NULL;
   return gf;
 }
 
-/*
- * Return an un-globbed name in the same form as if we were doing
- * globbing. 
- */
-char **
-make_name (path, section, name, cat)
-     register char *path;
-     register char *section;
-     register char *name;
-     register int cat;
-{
-  register int i = 0;
-  static char *names[3];
-  char buf[BUFSIZ];
-
-  if (cat)
-    sprintf (buf, "%s/cat%s/%s.%s", path, section, name, section);
-  else
-    sprintf (buf, "%s/man%s/%s.%s", path, section, name, section);
-
-  if (access (buf, R_OK) == 0)
-    names[i++] = strdup (buf);
-
-  /*
-   * If we're given a section that looks like `3f', we may want to try
-   * file names like .../man3/foo.3f as well.  This seems a bit
-   * kludgey to me, but what the hey...
-   */
-  if (section[1] != '\0')
-    {
-      if (cat)
-	sprintf (buf, "%s/cat%c/%s.%s", path, section[0], name, section);
-      else
-	sprintf (buf, "%s/man%c/%s.%s", path, section[0], name, section);
-
-      if (access (buf, R_OK) == 0)
-	names[i++] = strdup (buf);
-    }
-
-  names[i] = NULL;
-
-  return &names[0];
-}
-
-#ifdef DO_UNCOMPRESS
-char *
-get_expander (file)
-     char *file;
-{
-  char *expander = NULL;
-  int len = strlen (file);
-
-  if (file[len - 2] == '.')
-    {
-      switch (file[len - 1])
-	{
-#ifdef FCAT
-	case 'F':
-	  if (strcmp (FCAT, "") != 0)
-	    expander = strdup (FCAT);
-	  break;
-#endif
-#ifdef YCAT
-	case 'Y':
-	  if (strcmp (YCAT, "") != 0)
-	    expander = strdup (YCAT);
-	  break;
-#endif
-#ifdef ZCAT
-	case 'Z':
-	  if (strcmp (ZCAT, "") != 0)
-	    expander = strdup (ZCAT);
-	  break;
-#endif
-	default:
-	  break;
-	}
-    }
-  return expander;
-}
-#endif
-
 /*
  * Simply display the preformatted page.
  */
@@ -685,16 +434,12 @@ display_cat_file (file)
 
   if (access (file, R_OK) == 0)
     {
-#ifdef DO_UNCOMPRESS
       char *expander = get_expander (file);
 
-      if (expander != NULL)
+      if (expander != NULL && expander[0] != 0)
 	sprintf (command, "%s %s | %s", expander, file, pager);
       else
 	sprintf (command, "%s %s", pager, file);
-#else
-      sprintf (command, "%s %s", pager, file);
-#endif
 
       found = do_system_command (command);
     }
@@ -710,69 +455,109 @@ display_cat_file (file)
  * the input file name is returned.
  */
 char *
-ultimate_source (name, path)
-     char *name;
-     char *path;
+ultimate_source (name0)
+     char *name0;
 {
   FILE *fp;
-  char buf[BUFSIZ];
-  char ult[BUFSIZ];
+  char *name;
+  char *expander;
+  int expfl = 0;
+  char *fgr;
   char *beg;
   char *end;
+  char *cp;
+  char buf[BUFSIZ];
 
-  strcpy (ult, name);
-  strcpy (buf, name);
+/* TAMU FIX  -- this has to be static */
+  static char ultname[BUFSIZ];
 
- next:
+  name = name0;
 
-  if ((fp = fopen (ult, "r")) == NULL)
-    return buf;
+ again:
+  expander = get_expander (name);
+  if (expander && *expander) {
+      char command[BUFSIZ];
 
-  if (fgets (buf, BUFSIZ, fp) == NULL)
-    return ult;
+      sprintf (command, "%s %s", expander, name);
+      fp = popen (command, "r");
+      if (fp == NULL) {
+	  perror("popen");
+	  fprintf (stderr, "Attempt [%s] to expand man page failed\n",
+		   command);
+	  return (NULL);
+      }
+      fgr = fgets (buf, BUFSIZ, fp);
+      pclose (fp);
+      expfl = 1;
+  } else {
+      fp = fopen (name, "r");
+      if (fp == NULL && expfl) {
+	  char *extp = rindex (name0, '.');
+	  if (extp && *extp) {
+	      strcat(name, extp);
+	      fp = fopen (name, "r");
+	  }
+      }
+      if (fp == NULL) {
+	  perror("fopen");
+	  fprintf (stderr, "Cannot open man page %s\n", name);
+	  return (NULL);
+      }
+      fgr = fgets (buf, BUFSIZ, fp);
+      fclose (fp);
+  }
 
-  if (strlen (buf) < 5)
-    return ult;
+  if (fgr == NULL) {
+      perror("fgets");
+      fprintf (stderr, "Error reading man page %s\n", name);
+      return (NULL);
+  }
 
-  beg = buf;
-  if (*beg++ == '.' && *beg++ == 's' && *beg++ == 'o')
-    {
-      while ((*beg == ' ' || *beg == '\t') && *beg != '\0')
-	beg++;
+  if (strncmp(buf, ".so", 3))
+    return (name);
 
-      end = beg;
-      while (*end != ' ' && *end != '\t' && *end != '\n' && *end != '\0')
-	end++;
+  beg = buf+3;
+  while (*beg == ' ' || *beg == '\t')
+    beg++;
 
-      *end = '\0';
+  end = beg;
+  while (*end != ' ' && *end != '\t' && *end != '\n' && *end != '\0')
+    end++;
 
-      strcpy (ult, path);
-      strcat (ult, "/");
-      strcat (ult, beg);
+  *end = '\0';
 
-      strcpy (buf, ult);
+  if (name != ultname) {
+      strcpy(ultname, name);
+      name = ultname;
+  }
 
-      goto next;
-    }
+  /* If name ends in path/manx/foo.9x then use path, otherwise
+     try same directory. */
+  if ((cp = rindex(name, '/')) == NULL) /* very strange ... */
+    return 0;
+  *cp = 0;
+  if((cp = rindex(name, '/')) != NULL && !strncmp(cp+1, "man", 3))
+    strcpy(cp+1, beg);
+  else if((cp = rindex(beg, '/')) != NULL)
+    strcat(name, cp);
+  else {
+      strcat(name, "/");
+      strcat(name, beg);
+  }
 
-  if (debug)
-    fprintf (stderr, "found ultimate source file %s\n", ult);
-
-  return ult;
+  goto again;
 }
 
-void
-add_directive (first, d, file, buf)
-     int *first;
+static void
+add_directive (d, file, buf)
      char *d;
      char *file;
      char *buf;
 {
-  if (strcmp (d, "") != 0)
+  if ((d = getval(d)) != 0)
     {
-      if (*first)
+      if (*buf == 0)
 	{
-	  *first = 0;
 	  strcpy (buf, d);
 	  strcat (buf, " ");
 	  strcat (buf, file);
@@ -785,15 +570,15 @@ add_directive (first, d, file, buf)
     }
 }
 
-int
+static int
 parse_roff_directive (cp, file, buf)
   char *cp;
   char *file;
   char *buf;
 {
   char c;
-  int first = 1;
   int tbl_found = 0;
+  char *expander;
 
   while ((c = *cp++) != '\0')
     {
@@ -803,55 +588,43 @@ parse_roff_directive (cp, file, buf)
 
 	  if (debug)
 	    fprintf (stderr, "found eqn(1) directive\n");
-
-	  if (troff)
-	    add_directive (&first, EQN, file, buf);
-	  else
-	    add_directive (&first, NEQN, file, buf);
-
+	  add_directive ((do_troff ? "EQN" : "NEQN"), file, buf);
 	  break;
 
 	case 'g':
 
 	  if (debug)
 	    fprintf (stderr, "found grap(1) directive\n");
-
-	  add_directive (&first, GRAP, file, buf);
-
+	  add_directive ("GRAP", file, buf);
 	  break;
 
 	case 'p':
 
 	  if (debug)
 	    fprintf (stderr, "found pic(1) directive\n");
-
-	  add_directive (&first, PIC, file, buf);
-
+	  add_directive ("PIC", file, buf);
 	  break;
 
 	case 't':
 
 	  if (debug)
 	    fprintf (stderr, "found tbl(1) directive\n");
-
 	  tbl_found++;
-	  add_directive (&first, TBL, file, buf);
+	  add_directive ("TBL", file, buf);
 	  break;
 
 	case 'v':
 
 	  if (debug)
 	    fprintf (stderr, "found vgrind(1) directive\n");
-
-	  add_directive (&first, VGRIND, file, buf);
+	  add_directive ("VGRIND", file, buf);
 	  break;
 
 	case 'r':
 
 	  if (debug)
 	    fprintf (stderr, "found refer(1) directive\n");
-
-	  add_directive (&first, REFER, file, buf);
+	  add_directive ("REFER", file, buf);
 	  break;
 
 	case ' ':
@@ -868,27 +641,16 @@ parse_roff_directive (cp, file, buf)
 
  done:
 
-  if (first)
+  if (*buf == 0)
     return 1;
 
-#ifdef HAS_TROFF
-  if (troff)
-    {
-      strcat (buf, " | ");
-      strcat (buf, TROFF);
-    }
-  else
-#endif
-    {
-      strcat (buf, " | ");
-      strcat (buf, NROFF);
-    }
+  strcat (buf, " | ");
+  strcat (buf, getval(do_troff ? "TROFF" : "NROFF"));
 
-  if (tbl_found && !troff && strcmp (COL, "") != 0)
-    {
+  if (tbl_found && !do_troff && getval("COL") != 0) {
       strcat (buf, " | ");
-      strcat (buf, COL);
-    }
+      strcat (buf, getval("COL"));
+  }
 
   return 0;
 }
@@ -901,7 +663,13 @@ make_roff_command (file)
   char line [BUFSIZ];
   static char buf [BUFSIZ];
   int status;
-  char *cp;
+  char *cp, *tblpath, *colpath, *roffpath, *expander, *fgr;
+
+  expander = get_expander (file);
+  if (expander && *expander)
+    sprintf (buf, "%s %s", expander, file);
+  else
+    buf[0] = 0;
 
   if (roff_directive != NULL)
     {
@@ -917,38 +685,51 @@ make_roff_command (file)
 	gripe_roff_command_from_command_line (file);
     }
 
-  if ((fp = fopen (file, "r")) != NULL)
-    {
-      cp = &line[0];
-      fgets (line, 100, fp);
-      if (*cp++ == '\'' && *cp++ == '\\' && *cp++ == '"' && *cp++ == ' ')
-	{
-	  if (debug)
-	    fprintf (stderr, "parsing directive from file\n");
+  if (expander && *expander) {
+      char command[BUFSIZ];
 
-	  status = parse_roff_directive (cp, file, buf);
+      sprintf (command, "%s %s", expander, file);
+      fp = popen (command, "r");
+      if (fp == NULL) {
+	  perror("popen");
+	  fprintf (stderr, "Attempt [%s] to expand man page failed\n",
+		   command);
+	  return (NULL);
+      }
+      fgr = fgets (line, BUFSIZ, fp);
+      pclose (fp);
+  } else {
+      fp = fopen (file, "r");
+      if (fp == NULL) {
+	  perror("fopen");
+	  fprintf (stderr, "Cannot open man page %s\n", file);
+	  return (NULL);
+      }
+      fgr = fgets (line, BUFSIZ, fp);
+      fclose (fp);
+  }
 
-	  fclose (fp);
+  if (fgr == NULL) {
+      perror("fgets");
+      fprintf (stderr, "Error reading man page %s\n", file);
+      return (NULL);
+  }
 
-	  if (status == 0)
-	    return buf;
+  cp = &line[0];
+  if (*cp++ == '\'' && *cp++ == '\\' && *cp++ == '"' && *cp++ == ' ') {
+      if (debug)
+	fprintf (stderr, "parsing directive from file %s\n", file);
 
-	  if (status == -1)
-	    gripe_roff_command_from_file (file);
-	}
-    }
-  else
-    {
-      /*
-       * Is there really any point in continuing to look for
-       * preprocessor options if we can't even read the man page source? 
-       */
-      gripe_reading_man_file (file);
-      return NULL;
-    }
+      status = parse_roff_directive (cp, file, buf);
 
-  if ((cp = getenv ("MANROFFSEQ")) != NULL)
-    {
+      if (status == 0)
+	return buf;
+
+      if (status == -1)
+	gripe_roff_command_from_file (file);
+  }
+
+  if ((cp = getenv ("MANROFFSEQ")) != NULL) {
       if (debug)
 	fprintf (stderr, "parsing directive from environment\n");
 
@@ -959,53 +740,13 @@ make_roff_command (file)
 
       if (status == -1)
 	gripe_roff_command_from_env ();
-    }
+  }
 
   if (debug)
     fprintf (stderr, "using default preprocessor sequence\n");
 
-#ifdef HAS_TROFF
-  if (troff)
-    {
-      if (strcmp (TBL, "") != 0)
-	{
-	  strcpy (buf, TBL);
-	  strcat (buf, " ");
-	  strcat (buf, file);
-	  strcat (buf, " | ");
-	  strcat (buf, TROFF);
-	}
-      else
-	{
-	  strcpy (buf, TROFF);
-	  strcat (buf, " ");
-	  strcat (buf, file);
-	}
-    }
-  else
-#endif
-    {
-      if (strcmp (TBL, "") != 0)
-	{
-	  strcpy (buf, TBL);
-	  strcat (buf, " ");
-	  strcat (buf, file);
-	  strcat (buf, " | ");
-	  strcat (buf, NROFF);
-	}
-      else
-	{
-	  strcpy (buf, NROFF);
-	  strcat (buf, " ");
-	  strcat (buf, file);
-	}
+  (void) parse_roff_directive ("t", file, buf);
 
-      if (strcmp (COL, "") != 0)
-	{
-	  strcat (buf, " | ");
-	  strcat (buf, COL);
-	}
-    }
   return buf;
 }
 
@@ -1013,7 +754,7 @@ make_roff_command (file)
  * Try to format the man page and create a new formatted file.  Return
  * 1 for success and 0 for failure.
  */
-int
+static int
 make_cat_file (path, man_file, cat_file)
      register char *path;
      register char *man_file;
@@ -1035,8 +776,12 @@ make_cat_file (path, man_file, cat_file)
 	return 0;
       else
 #ifdef DO_COMPRESS
+	/* The cd is necessary, because of .so commands,
+	   like .so man1/bash.1 in bash_builtins.1.
+	   But it changes the meaning of man_file and cat_file,
+	   if these are not absolute. */
 	sprintf (command, "(cd %s ; %s | %s > %s)", path,
-		 roff_command, COMPRESSOR, cat_file);
+		 roff_command, getval("COMPRESS"), cat_file);
 #else
         sprintf (command, "(cd %s ; %s > %s)", path,
 		 roff_command, cat_file);
@@ -1057,7 +802,7 @@ make_cat_file (path, man_file, cat_file)
 	  chmod (cat_file, mode);
 
 	  if (debug)
-	    fprintf (stderr, "mode of %s is now %o\n", cat_file, mode);
+	    fprintf (stderr, "changed mode of %s to %o\n", cat_file, mode);
 	}
 
       signal (SIGINT, SIG_DFL);
@@ -1073,19 +818,32 @@ make_cat_file (path, man_file, cat_file)
     }
 }
 
+static int
+display_man_file(path, man_file)
+     char *path, *man_file;
+{
+  char *roff_command;
+  char command[BUFSIZ];
+
+  roff_command = make_roff_command (man_file, 0);
+  if (roff_command == NULL)
+    return 0;
+  else
+    sprintf (command, "(cd %s ; %s | %s)", path,
+	     roff_command, pager);
+
+  return do_system_command (command);
+}
+
 /*
  * Try to format the man page source and save it, then display it.  If
  * that's not possible, try to format the man page source and display
  * it directly.
- *
- * Note that we've already been handed the name of the ultimate source
- * file at this point.
  */
-int
-format_and_display (path, man_file, cat_file)
+static int
+format_and_display (path, man_file)
      register char *path;
      register char *man_file;
-     register char *cat_file;
 {
   int status;
   register int found;
@@ -1096,27 +854,47 @@ format_and_display (path, man_file, cat_file)
 
   if (access (man_file, R_OK) != 0)
     return 0;
-  
-  if (troff)
-    {
+
+  /* first test for contents  .so man1/xyzzy.1  */
+  /* (in that case we do not want to make a cat file identical
+     to cat1/xyzzy.1) */
+  man_file = ultimate_source (man_file);
+  if (man_file == NULL)
+    return 0;
+
+  if (do_troff) {
       roff_command = make_roff_command (man_file, 1);
       if (roff_command == NULL)
 	return 0;
       else
 	sprintf (command, "(cd %s ; %s)", path, roff_command);
 
-      found = do_system_command (command);
-    }
-  else
-    {
+      return do_system_command (command);
+  } else {
+      char *cat_file = convert_name (man_file);
+
+      if (debug)
+	fprintf (stderr, "will try to write %s if needed\n", cat_file);
+
+      /*
+       * If cat_file exists, check whether it is more recent.
+       * Otherwise, check for other cat files (maybe there are
+       * old .Z files that should be removed).
+       */
+
       status = is_newer (man_file, cat_file);
       if (debug)
-	fprintf (stderr, "status from is_newer() = %d\n");
+	fprintf (stderr, "status from is_newer() = %d\n", status);
+      if (status == -1 || status == -3) {
+	  /* what? man_file does not exist anymore? */
+	  fprintf (stderr, "format_and_display: stat %s failed\n", man_file);
+	  return(0);
+      }
 
-      if (status == 1 || status == -2)
-	{
+      if (status != 0) {
 	  /*
-	   * Cat file is out of date.  Try to format and save it.
+	   * Cat file is out of date (status = 1) or does not exist
+	   * (status = -2).  Try to format and save it.
 	   */
 	  if (print_where)
 	    {
@@ -1126,6 +904,7 @@ format_and_display (path, man_file, cat_file)
 	  else
 	    {
 	      found = make_cat_file (path, man_file, cat_file);
+
 #ifdef SECURE_MAN_UID
 	      if (!found)
 		{
@@ -1151,6 +930,12 @@ format_and_display (path, man_file, cat_file)
 		   * Creating the cat file worked.  Now just display it.
 		   */
 		  (void) display_cat_file (cat_file);
+
+		  /*
+		   * If we just created this cat file, unlink any others.
+		   */
+		  if (status == -2)
+		    remove_other_catfiles(cat_file);
 		}
 	      else
 		{
@@ -1158,14 +943,7 @@ format_and_display (path, man_file, cat_file)
 		   * Couldn't create cat file.  Just format it and
 		   * display it through the pager. 
 		   */
-		  roff_command = make_roff_command (man_file, 0);
-		  if (roff_command == NULL)
-		    return 0;
-		  else
-		    sprintf (command, "(cd %s ; %s | %s)", path,
-			     roff_command, pager);
-
-		  found = do_system_command (command);
+		    found = display_man_file (path, man_file);
 		}
 	    }
 	}
@@ -1194,11 +972,10 @@ format_and_display (path, man_file, cat_file)
  * section.
  */
 int
-try_section (path, section, name, glob)
+try_section (path, section, name)
      register char *path;
      register char *section;
      register char *name;
-     register int glob;
 {
   register int found = 0;
   register int to_cat;
@@ -1207,79 +984,45 @@ try_section (path, section, name, glob)
   register char **np;
 
   if (debug)
-    {
-      if (glob)
-	fprintf (stderr, "trying section %s with globbing\n", section);
-      else
-	fprintf (stderr, "trying section %s without globbing\n", section);
-    }
+    fprintf (stderr, "trying section %s\n", section);
 
-#ifndef NROFF_MISSING
-  /*
-   * Look for man page source files.
-   */
-  cat = 0;
-  if (glob)
-    names = glob_for_file (path, section, name, cat);
-  else
-    names = make_name (path, section, name, cat);
+  names = NULL;
 
-  if (names == (char **) -1 || *names == NULL)
+  if (*getval("NROFF")) {
     /*
-     * No files match.  See if there's a preformatted page around that
-     * we can display. 
+     * Look for man page source files.
      */
-#endif /* NROFF_MISSING */
-    {
-      if (!troff)
-	{
-	  cat = 1;
-	  if (glob)
-	    names = glob_for_file (path, section, name, cat);
-	  else
-	    names = make_name (path, section, name, cat);
+      cat = 0;
+      names = glob_for_file (path, section, name, cat);
+  }
 
-	  if (names != (char **) -1 && *names != NULL)
-	    {
-	      for (np = names; *np != NULL; np++)
-		{
-		  if (print_where)
-		    {
+  if (names == NULL || *names == NULL) {
+      /*
+       * No files match.  See if there's a preformatted page around
+       * that we can display. 
+       */
+
+      if (!do_troff) {
+	  cat = 1;
+	  names = glob_for_file (path, section, name, cat);
+
+	  if (names != NULL) {
+	      for (np = names; *np != NULL; np++) {
+		  if (print_where) {
 		      printf ("%s\n", *np);
 		      found++;
-		    }
-		  else
-		    {
+		  } else {
 		      found += display_cat_file (*np);
-		    }
-		}
-	    }
-	}
-    }
-#ifndef NROFF_MISSING
-  else
-    {
-      for (np = names; *np != NULL; np++)
-	{
-	  register char *cat_file = NULL;
-	  register char *man_file;
+		  }
+	      }
+	  }
+      }
+  } else {
+      for (np = names; *np != NULL; np++) {
+	  found += format_and_display (path, *np);
+      }
+  }
 
-	  man_file = ultimate_source (*np, path);
-
-	  if (!troff)
-	    {
-	      to_cat = 1;
-
-	      cat_file = convert_name (man_file, to_cat);
-
-	      if (debug)
-		fprintf (stderr, "will try to write %s if needed\n", cat_file);
-	    }
-
-	  found += format_and_display (path, man_file, cat_file);
-	}
-    }
-#endif /* NROFF_MISSING */
   return found;
 }
 
@@ -1296,52 +1039,77 @@ try_section (path, section, name, glob)
  * If preformatted pages are not supported, or the troff option is
  * being used, only look for the man page source file.
  *
+ * Note that globbing is necessary also if the section is given,
+ * since a preformatted man page might be compressed.
+ *
  */
 int
 man (name)
      char *name;
 {
   register int found;
-  register int glob;
-  register char **mp;
   register char **sp;
+  FILE *fp;
 
   found = 0;
 
+  /* allow  man ./manpage  for formatting explicitly given man pages */
+  if (index(name, '/') && (fp = fopen(name, "r"))) {
+      char fullname[BUFSIZ];
+      char fullpath[BUFSIZ];
+      char *path;
+      char *cp;
+
+      fclose (fp);
+      if (*name != '/' && getcwd(fullname, sizeof(fullname))) {
+	  strcat (fullname, "/");
+	  strcat (fullname, name);
+      } else
+	strcpy (fullname, name);
+
+      strcpy (fullpath, fullname);
+      if ((cp = rindex(fullpath, '/')) != NULL) {
+	  strcpy(cp+1, "..");
+	  path = fullpath;
+      } else
+	path = ".";
+
+      name = ultimate_source (fullname);
+      
+      return display_man_file (path, name);
+  }
+
   fflush (stdout);
   if (section != NULL)
-    {
-      for (mp = manpathlist; *mp != NULL; mp++)
-	{
-	  if (debug)
-	    fprintf (stderr, "\nsearching in %s\n", *mp);
+    found = man_from_section (section, name);
+  else {
+      for (sp = section_list; *sp != NULL; sp++) {
+	  found += man_from_section (*sp, name);
+	  if (found && !findall)
+	    break;
+      }
+  }
+  return found;
+}
 
-	  glob = 0;
+int
+man_from_section (sec, name)
+     char *sec, *name;
+{
+  register int found;
+  register struct dir *dp;
 
-	  found += try_section (*mp, section, name, glob);
+  found = 0;
 
-	  if (found && !findall)   /* i.e. only do this section... */
-	    return found;
-	}
-    }
-  else
-    {
-      for (sp = section_list; *sp != NULL; sp++)
-	{
-	  for (mp = manpathlist; *mp != NULL; mp++)
-	    {
-	      if (debug)
-		fprintf (stderr, "\nsearching in %s\n", *mp);
+  for (dp = srchlist.nxt; dp; dp = dp->nxt) {
+      if (debug)
+	fprintf (stderr, "\nsearching in %s\n", dp->dir);
 
-	      glob = 1;
+      found += try_section (dp->dir, sec, name);
 
-	      found += try_section (*mp, *sp, name, glob);
-
-	      if (found && !findall)   /* i.e. only do this section... */
-		return found;
-	    }
-	}
-    }
+      if (found && !findall)   /* i.e. only do this section... */
+	return found;
+  }
   return found;
 }
 
@@ -1373,7 +1141,7 @@ get_section_list ()
 
       tmp_section_list[i++] = strdup (p);
 
-      if (end == NULL)
+      if (end == NULL || i+1 == sizeof(tmp_section_list))
 	break;
     }
 
